/*

 * formula.cpp
 *
 *  Created on: Jan 14, 2009
 *      Author: qwan
 */

#include "formula.h"
#include "common.h"


//extern CFormulaNode** matrix;		//[no_package][no_table]

void errlog(char* s){
	printf("error in %s, function %s, line %d, message %s\n", __FILE__, __FUNCTION__, __LINE__,s);
}

int CFormulaNode::cnt = 0;
int CSetformula::cnt = 0;

CFormulaNode* node_intersection(CFormulaNode* set1, CFormulaNode* set2){

	CFormulaNode* node = new CFormulaNode;

	if(set1->range != set2->range){
//		errlog("error");
		return 0;
	}
	else if(set1->set == NULL || set2->set == NULL){
//		errlog("error");
		return 0;
	}
	else if(set1->isfull()){
		node->copy(*set2);
		return node;
	}
	else if(set2->isfull()){
		node->copy(*set1);
		return node;
	}

	Pt_Collection* result = new Pt_Collection;

	if(!set1->isComplement && !set2->isComplement){
		::set_intersection(set1->set->begin(), set1->set->end(), set2->set->begin(), set2->set->end(), ::inserter(*result, (*result).begin()));
		node->isComplement = false;
	}
	else if(!set1->isComplement&&set2->isComplement){
		::set_difference(set1->set->begin(), set1->set->end(), set2->set->begin(), set2->set->end(), ::inserter(*result, (*result).begin()));
		node->isComplement = false;
	}
	else if(set1->isComplement&&set2->isComplement){
			::set_union(set1->set->begin(), set1->set->end(), set2->set->begin(), set2->set->end(), ::inserter(*result, (*result).begin()));
			node->isComplement = true;
	}
	else if(set1->isComplement&&!set2->isComplement){
		::set_difference(set2->set->begin(), set2->set->end(), set1->set->begin(), set1->set->end(), ::inserter(*result, (*result).begin()));
		//Pt_Collection* tp = new Pt_Collection;
		//::set_intersection(set2->set->begin(), set2->set->end(), result->begin(), result->end(), ::inserter(*tp, (*tp).begin()));
//		delete result;
//		result = tp;
		node->isComplement = false;
	}

	node->range = set1->range;
	node->set = result;

	if(node->set->size() == 0 && !node->isComplement){
		delete node;
		return 0;
	}
	else if(node->set->size() == (unsigned int)node->range && node->isComplement)
	{
		delete node;
		return 0;
	}
	else{
		return node;
	}


}


/*
void CSetformula::init(int n, int p, CFormulaNode** matrix){

	dim = n;
	CFormulaNode* tp = new CFormulaNode[n];

	int nsq = 1<<n;	//2^n, n < 31, to be effective n < 10
	for(int i = 1; i < nsq; i++){
		unsigned int s = i;
		int insertNeed = true;

		for(int j = 0; j < n; j++){
			tp[j].copy(matrix[p][j]);
			if((s & 1) == 0){
				tp[j].isComplement = false;
				if(tp[j].set->size() == 0){
					insertNeed = false;
				}
			}
			else
			{
				tp[j].isComplement = true;
				if(tp[j].set->size() == tp[j].range){
					insertNeed = false;
				}
			}
			s=s>>1;
		}
		if(insertNeed){
			formula.push_back(tp);
			tp = new CFormulaNode[n];
		}

	}


	delete[] tp;
};
*/

void CSetformula::init(int n, int p, CFormulaNode** matrix){

	dim = n;
	CFormulaNode* tp = new CFormulaNode[n];

//	int nsq = 1<<n;	//2^n, n < 31, to be effective n < 10

	int base = rand() % n;

	for(int k = 0; k < n; k++){

		int i = (base + k) % n;

		int insertNeed = true;

		for(int j=0;j<i;j++){
			tp[j].copy(matrix[p][j]);
			if(tp[j].size()==0){
				insertNeed = false;
			}
		}

		tp[i].copy(matrix[p][i]);
		tp[i].isComplement = true;

		if(tp[i].size()==0){
			insertNeed = false;
		}

		for(int j=i+1;j<n;j++){

			if(tp[j].set == NULL){
				tp[j].set = new Pt_Collection();
			}
			else{
				tp[j].set->clear();
			}

			tp[j].range = matrix[p][j].range;
			tp[j].isComplement = true;

		}

		if(insertNeed){
			formula.push_back(tp);
			tp = new CFormulaNode[n];
		}

	}


	delete[] tp;
};


void CSetformula::show(){
	printf("\ndim is %d", dim);
	printf("\n============--------show node---------============");
	printf("\ncontains %d", formula.size());

	list<CFormulaNode* >::iterator i;
	for(i = formula.begin();i != formula.end(); i++){
		for(int j =0 ;j < dim; j++)
			(*i)[j].show();

		printf("\n========----------------------=============\n");
	}

}

CSetformula* formula_intersection(CSetformula* f1, CSetformula* f2){

	CSetformula* result = new CSetformula;
	list<CFormulaNode* >::iterator i;
	list<CFormulaNode* >::iterator j;
	result->dim = f1->dim;

	if(f1->dim!=f2->dim){
		//::errlog("error");
	}

	CFormulaNode** re = new CFormulaNode*[f1->dim];

	for(int k=0;k<f1->dim;k++){
		re[k] = NULL;
	}

	for(i = f1->formula.begin(); i!=f1->formula.end(); i++){
		for(j = f2->formula.begin(); j!=f2->formula.end(); j++){
			CFormulaNode* p1 = *i;
			CFormulaNode* p2 = *j;

			bool insertNeed = true;

			for(int k = 0; k < f1->dim; k++)
			{
				if(insertNeed){

					CFormulaNode* presult = node_intersection(&p1[k], &p2[k]);

					if(presult == NULL){
						for(int m=0;m<k;m++){
							delete re[m];
							re[m] = NULL;
						}
						insertNeed = false;
						break;
					}
					else
					{
						if(re[k]!= NULL){
							printf("mem error!!!!!");
							delete re[k];
						}
						re[k] = presult;
					}

				}
			}

			if(insertNeed){
				CFormulaNode* nodeArray = new CFormulaNode[f1->dim];
				for(int k=0;k<f1->dim;k++){
					nodeArray[k].copy(*re[k]);
					delete re[k];
					re[k] = NULL;
				}
				result->formula.push_back(nodeArray);
			}

		}
	}

	delete[] re;

	return result;

};


//else if(presult->size() == 0){
//	delete presult;
//	for(int m=0;m<k;m++){
//		delete re[m];
//		re[m] = NULL;
//	}
//	insertNeed = false;
//	break;
//
//}

int CSetformula::size(){
	int sum = 0;
	for(list<CFormulaNode*>::iterator i =formula.begin(); i!=formula.end();i++){
		CFormulaNode* array = *i;
		int m = 1;
		for(int j=0;j<dim;j++)
		{
			m*=array[j].size();
		}
		if(m == 0){
			printf("error in CSetformula::check()");
		}

		sum+=m;
	}

	return sum;
}
